package be.iminds.ilabt.jfed.testing.tests.highlevel;

import be.iminds.ilabt.jfed.*;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentController;
import be.iminds.ilabt.jfed.experiment.ExperimentControllerFactory;
import be.iminds.ilabt.jfed.experiment.ExperimentState;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.HighLevelModule;
import be.iminds.ilabt.jfed.highlevel.SfaOnlyExperimentModule;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartControllerManager;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.log.cache.ApiCallDetailsCache;
import be.iminds.ilabt.jfed.lowlevel.ApiInfo;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.api.test.CommonAMTest;
import be.iminds.ilabt.jfed.lowlevel.api.test.NodeLoginTester;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.StringRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.testing.base.ApiTest;
import be.iminds.ilabt.jfed.testing.base.ApiTestMetaData;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.KeyUtil;
import be.iminds.ilabt.jfed.util.TargetAuthority;
import be.iminds.ilabt.jfed.util.TextUtil;
import com.google.common.eventbus.EventBus;
import com.google.inject.AbstractModule;
import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Module;
import com.google.inject.util.Modules;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.inject.Inject;
import java.io.IOException;
import java.io.PrintStream;
import java.security.NoSuchAlgorithmException;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * GuiLogicTest
 */
public class GuiLogicTest extends ApiTest {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(GuiLogicTest.class);

    @Nonnull Server userAuthorityServer;

    private final Injector injector;

    @Inject
    public GuiLogicTest(@Nonnull Injector injector,
                        @Nonnull Logger logger,
                        @Nonnull TargetAuthority testedAuthority,
                        @Nonnull GeniUserProvider geniUserProvider,
                        @Nonnull JFedConnectionProvider connectionProvider,
                        @Nonnull TestbedInfoSource testbedInfoSource,
                        @Nonnull AuthorityFinder authorityFinder,
                        @Nonnull JFedPreferences jFedPreferences) {
        super(logger,
                testedAuthority, geniUserProvider,
                connectionProvider, testbedInfoSource,
                authorityFinder, jFedPreferences);

        this.injector = injector;

        if (user == null)
            throw new IllegalStateException("User is null");

        Server uas = user.getUserAuthorityServer();
        if (uas == null)
            throw new IllegalStateException("User has no Authority Server");

        userAuthorityServer = uas;
    }
    private static final ApiTestMetaData metadata = new ApiTestMetaData() {
        @Nonnull
        @Override
        public String getTestDescription() {
            return "A test using the jFed GUI Editor logic. This will reserve resources, test login, and delete them again.";
        }

        @Nonnull
        @Override
        public List<String> getReqConfKeys() {
            return Collections.emptyList();
        }

        @Nonnull
        @Override
        public List<String> getOptConfKeys() {
            List<String> res = new ArrayList<>();
            return res;
        }
    };
    public static ApiTestMetaData getMetaData() {
        return metadata;
    }

    NodeLoginTester nodeLoginTester;

    @Override
    public void setUp() throws Exception {
        this.nodeLoginTester =
                new NodeLoginTester(this, null, null, null);
    }

    //    private String requestRspec = "<?xml version='1.0'?>\n" +
//            "<rspec xmlns=\"http://www.geni.net/resources/rspec/3\" type=\"request\"  xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd \">\n" +
//            "  <node client_id=\"node0\" exclusive=\"false\" component_manager_id=\"urn:publicid:IDN+docker.ilabt.iminds.be+authority+am\">\n" +
//            "    <sliver_type name=\"docker-container\"/>\n" +
//            "  </node>\n" +
//            "</rspec>";
    private String requestRspec = "<?xml version='1.0'?>\n" +
            "<rspec xmlns=\"http://www.geni.net/resources/rspec/3\" type=\"request\"  xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd \">\n" +
            "    <node client_id=\"node0\" exclusive=\"true\" component_manager_id=\"urn:publicid:IDN+wall1.ilabt.iminds.be+authority+cm\">\n" +
            "    <sliver_type name=\"raw-pc\"/>\n" +
            "  </node>\n" +
            "</rspec>";
    private int maxNodeNameLength = 5; //TODO: automatically get from rspec
    private List<Server> serversInRspec;
    private RequestRspecSource requestRspecSource;

    @Test(groups = {"init"})
    public void getRspec() throws GeniUrn.GeniUrnParseException {
        GeniUrn serverUrn = new GeniUrn("urn:publicid:IDN+docker.ilabt.iminds.be+authority+am");
        serversInRspec = Collections.singletonList(
                authorityFinder.findByUrn(
                        serverUrn,
                        AuthorityFinder.Purpose.REQUEST_RSPEC)
        );
        requestRspecSource = new RequestRspecSource(requestRspec, ModelRspecType.BASIC);
    }

    private UserAndSliceApiWrapper userAndSliceApiWrapper;
    private String project;
    @Test(hardDepends = {"getRspec"}, groups = {"init"})
    public void findProject() {
        userAndSliceApiWrapper =
                this.getAutomaticUserAndSliceApiWrapperFactory().create();

        project = CommonAMTest.findProject(
                this,
                this.getLogger(),
                null, /*todo: get from config*/
                user,
                user.getUserUrn(),
                userAndSliceApiWrapper,
                true);

        if (project != null) {
            note("Will use project: \"" + project+"\"");
        } else {
            note("Will NOT use a project");
        }
    }

    private Experiment experiment;
    @Test(hardDepends = {"getRspec", "findProject"}, groups = {"init"})
    public void createExperiment() {
        String sliceName = CommonAMTest.createSliceName(
                this,
                userAuthorityServer,
                ApiInfo.getDefaultComponentManagerForEach(serversInRspec),
                "g",
                maxNodeNameLength,
                null);

        note("Will create slice \""+sliceName+"\"");

        Instant requestedEndTime = Instant.now().plus(30, ChronoUnit.MINUTES); //TODO: make configurable
        note("Created slice (and slivers) will expire at "+requestedEndTime+"");

//        List<String> sshKeys = Collections.singletonList(KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(user.getPublicKey()));
        List<String> sshKeys = Collections.singletonList(nodeLoginTester.getSshKeyHelper().getSshPublicKeyString());
        UserSpec userSpec = new UserSpec(user.getUserUrnString(), sshKeys);
        List<UserSpec> userSpecs = Collections.singletonList(userSpec);

        experiment = new Experiment(
                sliceName,
                project,
                requestRspecSource,
                null,
                requestedEndTime,
                userSpecs,
                false);
    }

    private Injector highLevelInjector;
    @Test(hardDepends = {"getRspec"}, groups = {"run"})
    public void createInjector() {
        List<com.google.inject.AbstractModule> extraMods = new ArrayList<>();

        boolean hasHighLevelModule = false;
        boolean hasSfaOnlyExperimentModule = false;

        try {
            hasHighLevelModule = injector.getInstance(ApiCallDetailsCache.class) != null;
        } catch (Exception e) {
            LOG.info("Binding from HighLevelModule does not exists "+e.getClass().getName(), e);
        }
        try {
            hasSfaOnlyExperimentModule = injector.getInstance(ExperimentPartControllerManager.class) != null;
        } catch (Exception e) {
            LOG.info("Binding from SfaOnlyExperimentModule does not exists "+e.getClass().getName(), e);
        }

        if (!hasHighLevelModule) {
            extraMods.add(new HighLevelModule());
            note("Will add HighLevelModule to injector");
        } else {
            note("Will NOT add HighLevelModule to injector");
        }
        if (!hasSfaOnlyExperimentModule) {
            extraMods.add(new SfaOnlyExperimentModule());
            note("Will add SfaOnlyExperimentModule to injector");
        } else {
            note("Will NOT add SfaOnlyExperimentModule to injector");
        }

//        AbstractModule loggerOverride = new AbstractModule() {
//            @Override
//            protected void configure() {
//                todo;
//            }
//        };
//
//        todo; Modules.override(extraMods).with(loggerOverride);

        highLevelInjector = injector.createChildInjector();
        assertNotNull(highLevelInjector);
        note("Created injector");

        Logger injectorLogger = injector.getInstance(Logger.class);
        //noinspection ObjectEquality
        if (getLogger() != injectorLogger) {
            errorFatal("Invalid logger in injector");
        } else {
            note("Logger OK");
        }
    }

    private boolean notEnoughResourcesDetected;
    private ExperimentController experimentController;
    private long maxTimeToWaitForReadyInMinutes = 5;
    @Test(hardDepends = {"createExperiment", "createInjector"}, groups = {"run"})
    public void runExperiment() {
        //TODO later, we want to merge createExperiment and runExperiment and replace it by:
//        StartExperimentEvent startExperimentEvent =
//                new StartExperimentEvent(
//                        sliceName,
//                        project,
//                        controller.getExpirationTime(),
//                        requestRspecSource,
//                        controller.isIncludeSshKeys(),
//                        controller.isRequestSla());
//        eventBus.post(startExperimentEvent);

        notEnoughResourcesDetected = false;

        final ExperimentControllerFactory experimentControllerFactory =
                highLevelInjector.getInstance(ExperimentControllerFactory.class);
        assertNotNull(experimentControllerFactory);

        experimentController =
                experimentControllerFactory.createExperimentController(experiment);
        assertNotNull(experimentController);

//        experimentController.setExperimentTester(experimenterJobFactory::createTestConnectivityJob);
//        experimentController.addListener(bugReportExperimentControllerListener);

        experimentController.start();
        note("Started experimentController. ExperimentState is now " + experiment.getExperimentState());

        //not seperate methods: we might miss calls in between methods otherwise.
//    }
//
//
//    @Test(hardDepends = {"runExperiment"}, groups = {"run"})
//    public void waitForReady() {

        long now = System.currentTimeMillis();
        long deadline = now + (maxTimeToWaitForReadyInMinutes * 60 * 1000L);
        while (now < deadline) {
            note("Current ExperimentState=" + experiment.getExperimentState());
            note("Checking if experiment is ready.");
            if (experiment.getExperimentState() == ExperimentState.READY) {
                note("Ready, because ExperimentState is READY");
                return;
            }
            if (experiment.getExperimentState().isTerminated()) {
                errorFatal("Failure, because ExperimentState is terminated");
                return;
            }
            if (!experiment.getExperimentState().isActiveState()) {
                //TODO detect and set notEnoughResourcesDetected
                errorFatal("Failure, because ExperimentState is not active state");
                return;
            }

            try {
                Thread.sleep(30000 /*ms*/);
            } catch (InterruptedException e) { /* Ignore*/ }
            now = System.currentTimeMillis();
        }

        note("Current ExperimentState=" + experiment.getExperimentState());
        errorFatal("Experiment did not become active on time.");
    }

    public boolean isValidGeniRspec(String rspec) {
        return new BasicStringRspec(rspec).isValidRspec();
    }

    boolean validManifest;

    @Test(hardDepends = {"runExperiment"}, groups = {"login"})
    public void checkManifest() throws JFedException {
        if (notEnoughResourcesDetected) {
            warn("Test skipped because not enough free resources detected while tying to create the sliver(s)");
            return;
        }

        Slice slice = experiment.getSlice();
        ManifestRspecSource manifestRspecSource = slice == null ? null : slice.getManifestRspec();
        String manifestRspec = manifestRspecSource == null ? null : manifestRspecSource.getRspecXmlString();

        assertNotNull(manifestRspec);
        boolean validRequest = isValidGeniRspec(requestRspec);
        validManifest = isValidGeniRspec(manifestRspec);

        setErrorsNotFatal();
        if (requestRspec != null && manifestRspec != null && validRequest && validManifest) {
            assertTrue(CommonAMTest.sameNodesInRequestAndManifest(this, requestRspec, manifestRspec));
        }

        setErrorsFatal();
        if (validManifest) {
            boolean foundNodeLogin = nodeLoginTester.parseSshInfoFromGeni3ManifestRspec(manifestRspec, null, null);
            if (foundNodeLogin)
                note("Successfully found node login in manifest");
            else
                note("Did not find node login info in manifest");

            if (TextUtil.objectToBoolean(getTestConfig().getProperty("skip_final_manifest_check")) != Boolean.TRUE)
                nodeLoginTester.checkManifestCorrectness(manifestRspec, testedAuthority);
        }
    }

    @Test(hardDepends = {"checkManifest"}, groups = {"login"})
    public void login() throws NoSuchAlgorithmException, IOException {
        if (notEnoughResourcesDetected) {
            warn("Test skipped because not enough free resources detected while tying to create the sliver(s)");
            return;
        }
        if (!validManifest) {
            skip("Test skipped because manifest is not a geni version 3 RSpec (so this test does not know how to extract node login info)");
            return;
        }

        assert nodeLoginTester != null;

        //TODO set proxy from config

        //TODO support nodelogin_force_ip_version
//        String nodeloginForceIpVersion = getTestConfig().getProperty("nodelogin_force_ip_version");
//        if (nodeloginForceIpVersion != null) {
//            try {
//                nodeLoginTester.setNodeLoginForceIpVersion(Integer.parseInt(nodeloginForceIpVersion));
//            } catch (NumberFormatException e) {
//                warn("nodelogin_force_ip_version property is set to invalid value: \""+nodeloginForceIpVersion+"\". Supported: 4 or 6");
//            }
//        }

//        Boolean nodeloginUseExternalSsh = TextUtil.objectToBoolean(getTestConfig().getProperty("nodelogin_use_external_ssh"));
//        if (nodeloginUseExternalSsh == null) nodeloginUseExternalSsh = false;

        boolean nodeloginUseExternalSsh = false;
        //noinspection ConstantConditions
        nodeLoginTester.testNodeLogin(false/*only note, no warning*/, !nodeloginUseExternalSsh);
    }

    private long maxTimeToWaitForDeleteInMinutes = 5;

    @Test(softDepends = {"login"},
            hardDepends = {"getRspec", "createInjector", "createExperiment"},
            groups = {"cleanup"})
    public void deleteExperiment() {
        note("Will call experimentController.stop();");
        experimentController.stop();

        note("Will wait until experiment is terminated");
        long now = System.currentTimeMillis();
        long deadline = now + (maxTimeToWaitForDeleteInMinutes * 60 * 1000L);
        while (now < deadline) {
            note("Current ExperimentState=" + experiment.getExperimentState());
            if (experiment.getExperimentState() == ExperimentState.EMPTY ||
                    experiment.getExperimentState() == ExperimentState.EXPIRED ||
                    experiment.getExperimentState() == ExperimentState.FAILED) {
                note("Experiment was terminated");
                return;
            }

            try {
                Thread.sleep(30000 /*ms*/);
            } catch (InterruptedException e) { /* Ignore*/ }
            now = System.currentTimeMillis();
        }

        note("Current ExperimentState=" + experiment.getExperimentState());
        errorFatal("Experiment did not terminate within "+maxTimeToWaitForDeleteInMinutes+" minutes. Gave up waiting for it.");
    }
}
